#include "tail.h"

void Usage(void)
{
  printf("tail %s created by Jorgen Bosman\n",TAIL_VERSION);
  printf("Usage: tail [options] file\n");
  printf("Options:\n");
  printf("  -[x]f : Follow the file and display new lines as they are written to the file\n");
  printf("          Optionally display the x last lines first (default=10)\n");
  printf("  -n x  : Displays the last x lines of the file\n");
  printf("  -s x  : Sleeps x miliseconds between 2 read attempts for -f option\n");
  exit(-1);
}

void DisplayLastError(void)
{
  LPVOID lpMsgBuf;
  
  FormatMessage( 
    FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
    NULL,
    GetLastError(),
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
    (LPTSTR) &lpMsgBuf,    0,    NULL );
  printf("%s\n",lpMsgBuf);
  LocalFree( lpMsgBuf );
}

__int64 myFileSeek(HANDLE hf,__int64 distance,DWORD MoveMethod)
{
  LARGE_INTEGER li;

  li.QuadPart = distance;

  li.LowPart = SetFilePointer (hf, li.LowPart, &li.HighPart, MoveMethod);

  if (li.LowPart == 0xFFFFFFFF && GetLastError() != NO_ERROR)
  {
    li.QuadPart = -1;
  }

  return li.QuadPart;
}



int Tail(char *filename,Options options)
{
  HANDLE f;
  __int64 pos=0,filesize=0;
  unsigned lines=0,x,oldpos=0;
  char buf[512];
  DWORD bytes_read=0;

  f=CreateFile(filename,
               GENERIC_READ,
               FILE_SHARE_DELETE|FILE_SHARE_READ|FILE_SHARE_WRITE,
               NULL,
               OPEN_EXISTING,
               FILE_ATTRIBUTE_NORMAL,
               NULL);
  if (f==INVALID_HANDLE_VALUE)
  {
    DisplayLastError();
    return -1;
  }
  filesize=pos=myFileSeek(f,0,FILE_END);
  while (lines<options.nlastlines && pos>(__int64)0)
  {
    pos--;
    myFileSeek(f,pos,FILE_BEGIN);
    if (!ReadFile(f,buf,1,&bytes_read,NULL))
    {
      DisplayLastError();
      CloseHandle(f);
      return -1;
    }
    if (buf[0]=='\n')
    {
      if (pos!=filesize-(__int64)1)
        lines++;
    }
  }
  if (pos==(__int64)0)
    myFileSeek(f,pos,FILE_BEGIN);
  for (;;)
  {
    if (!ReadFile(f,buf,511,&bytes_read,NULL))
    {
      DisplayLastError();
      CloseHandle(f);
      return -1;
    }    
    if (bytes_read==0)
    {
      if (options.follow_file)
      {
        CloseHandle(f);
        Sleep(options.sleeptime);
        f=CreateFile(filename,
                     GENERIC_READ,
                     FILE_SHARE_DELETE|FILE_SHARE_READ|FILE_SHARE_WRITE,
                     NULL,
                     OPEN_EXISTING,
                     FILE_ATTRIBUTE_NORMAL,
                     NULL);
        if (f==INVALID_HANDLE_VALUE)
        {
          DisplayLastError();
          return -1;
        }
        filesize=myFileSeek(f,0,FILE_END);
        if (filesize<pos)
          pos=filesize;
        myFileSeek(f,pos+(__int64)1,FILE_BEGIN);
      }
      else
        break;
    }
    pos+=(__int64)bytes_read;
    buf[bytes_read]='\0';
    oldpos=0;
    for (x=0;x<bytes_read;x++)
    {
      if (buf[x]=='\n')
      {
        buf[x]='\0';
        printf("%s\n",&buf[oldpos]);
        oldpos=x+1;
      }
    }
    printf("%s",&buf[oldpos]);
    fflush(stdout);
  }
  CloseHandle(f);
  return 0;
}

int main(int argc,char *argv[])
{
  int x;
  Options options={10,1000,0};
  char *filename=NULL;

  for (x=1;x<argc;x++)
  {
    if (argv[x][0]=='-')
    {
      if (!strcmp(argv[x],"-n"))
      {
        x++;
        if (x<argc)
          options.nlastlines=atoi(argv[x]);
        else
          Usage();
      }
      if (!strcmp(argv[x],"-s"))
      {
        x++;
        if (x<argc)
          options.sleeptime=atoi(argv[x]);
        else
          Usage();
      }
      else if (argv[x][strlen(argv[x])-1]=='f')
      {
        options.follow_file=1;
        if (strcmp(argv[x],"-f"))
        {
          argv[x][strlen(argv[x])-1]='\0';
          options.nlastlines=atoi(&argv[x][1]);
        }
      }
    }
    else if (filename==NULL)
      filename=strdup(argv[x]);
    else
      Usage();
  }
  if (filename==NULL)
    Usage();

  x=Tail(filename,options);

  free(filename);
  return x;
}